def get_colnames(score: str):

    colnames = {
        "charlson": {
            "aids": "AIDS or HIV",
            "ami": "acute myocardial infarction",
            "canc": "cancer any malignancy",
            "cevd": "cerebrovascular disease",
            "chf": "congestive heart failure",
            "copd": "chronic obstructive pulmonary disease",
            "dementia": "dementia",
            "diab": "diabetes without complications",
            "diabwc": "diabetes with complications",
            "hp": "hemiplegia or paraplegia",
            "metacanc": "metastatic solid tumour",
            "mld": "mild liver disease",
            "msld": "moderate or severe liver disease",
            "pud": "peptic ulcer disease",
            "pvd": "peripheral vascular disease",
            "rend": "renal disease",
            "rheumd": "rheumatoid disease",
        },
        "elixhauser": {
            "aids": " AIDS/HIV",
            "alcohol": " alcohol abuse",
            "blane": " blood loss anaemia",
            "carit": " cardiac arrhythmias",
            "chf": " congestive heart failure",
            "coag": " coagulopathy",
            "cpd": " chronic pulmonary disease",
            "dane": " deficiency anaemia",
            "depre": " depression",
            "diabc": " diabetes complicated",
            "diabunc": " diabetes uncomplicated",
            "drug": " drug abuse",
            "fed": " fluid and electrolyte disorders",
            "hypc": " hypertension complicated",
            "hypothy": " hypothyroidism",
            "hypunc": " hypertension uncomplicated",
            "ld": " liver disease",
            "lymph": " lymphoma",
            "metacanc": " metastatic cancer",
            "obes": " obesity",
            "ond": " other neurological disorders",
            "para": " paralysis",
            "pcd": " pulmonary circulation disorders",
            "psycho": " psychoses",
            "pud": " peptic ulcer disease excluding bleeding",
            "pvd": " peripheral vascular disorders",
            "rf": " renal failure",
            "rheumd": " rheumatoid arthritis/collaged vascular disease",
            "solidtum": " solid tumour without metastasis",
            "valv": " valvular disease",
            "wloss": " weight loss",
        },
    }

    return colnames[score]



import pandas as pd


def assignzero(df: pd.DataFrame, score: str):
    if "charlson" in score:
        # "Mild liver disease" (`mld`) and "Moderate/severe liver disease" (`msld`)
        # x[msld == 1, mld := 0]
        df.mld = df.mld.where(df.msld == 0, 0)

        # "Diabetes" (`diab`) and "Diabetes with complications" (`diabwc`)
        # x[diabwc == 1, diab := 0]
        df.diab = df.diab.where(df.diabwc == 0, 0)

        # "Cancer" (`canc`) and "Metastatic solid tumour" (`metacanc`)
        # x[metacanc == 1, canc := 0]
        df.canc = df.canc.where(df.metacanc == 0, 0)

    elif "elixhauser" in score:
        # "Hypertension, uncomplicated" (`hypunc`) and "Hypertension, complicated" (`hypc`)
        # x[hypc == 1, hypunc := 0]
        df.hypunc = df.hypunc.where(df.hypc == 0, 0)

        # "Diabetes, uncomplicated" (`diabunc`) and "Diabetes, complicated" (`diabc`)
        # x[diabc == 1, diabunc := 0]
        df.diabunc = df.diabunc.where(df.diabc == 0, 0)

        # "Solid tumour" (`solidtum`) and "Metastatic cancer" (`metacanc`)
        # x[metacanc == 1, solidtum := 0]
        df.solidtum = df.solidtum.where(df.metacanc == 0, 0)

    return df



# Empty list
from .mapping import mapping

weights = dict()

for w in mapping.keys():
    weights[w] = dict()
    if "charlson" in w:
        # Charlson-compatible weights:
        # Original Charlson weights
        weights[w]["charlson"] = dict(
            ami=1,
            chf=1,
            pvd=1,
            cevd=1,
            dementia=1,
            copd=1,
            rheumd=1,
            pud=1,
            mld=1,
            diab=1,
            diabwc=2,
            hp=2,
            rend=2,
            canc=2,
            msld=3,
            metacanc=6,
            aids=6,
        )
        # Quan (2011)
        weights[w]["quan"] = dict(
            ami=0,
            chf=2,
            pvd=0,
            cevd=0,
            dementia=2,
            copd=1,
            rheumd=1,
            pud=0,
            mld=2,
            diab=0,
            diabwc=1,
            hp=2,
            rend=1,
            canc=2,
            msld=4,
            metacanc=6,
            aids=2,
        )
        # UK SHMI (version 1.35 Sept 2020)
        weights[w]["shmi"] = dict(
            ami=5,
            chf=13,
            pvd=6,
            cevd=11,
            dementia=14,
            copd=4,
            rheumd=4,
            pud=9,
            mld=8,
            diab=3,
            diabwc=-1,
            hp=1,
            rend=10,
            canc=8,
            msld=18,
            metacanc=14,
            aids=2,
        )
        # UK SHMI (version 1.35 Sept 2020) - modified to increase weighting for diabwc from -1 to 4
        weights[w]["shmi_modified"] = dict(
            ami=5,
            chf=13,
            pvd=6,
            cevd=11,
            dementia=14,
            copd=4,
            rheumd=4,
            pud=9,
            mld=8,
            diab=3,
            diabwc=4,
            hp=1,
            rend=10,
            canc=8,
            msld=18,
            metacanc=14,
            aids=2,
        )
    else:
        # Elixhauser-compatible weights:
        # van Walraven
        weights[w]["vw"] = dict(
            chf=7,
            carit=5,
            valv=-1,
            pcd=4,
            pvd=2,
            hypunc=0,
            hypc=0,
            para=7,
            ond=6,
            cpd=3,
            diabunc=0,
            diabc=0,
            hypothy=0,
            rf=5,
            ld=11,
            pud=0,
            aids=0,
            lymph=9,
            metacanc=12,
            solidtum=4,
            rheumd=0,
            coag=3,
            obes=-4,
            wloss=6,
            fed=5,
            blane=-2,
            dane=-2,
            alcohol=0,
            drug=-7,
            psycho=0,
            depre=-3,
        )
        # Swiss weights
        weights[w]["swiss"] = dict(
            chf=13,
            carit=6,
            valv=-1,
            pcd=6,
            pvd=3,
            hypunc=-4,
            hypc=-3,
            para=11,
            ond=10,
            cpd=3,
            diabunc=1,
            diabc=-1,
            hypothy=-3,
            rf=8,
            ld=16,
            pud=0,
            aids=0,
            lymph=9,
            metacanc=17,
            solidtum=10,
            rheumd=-1,
            coag=9,
            obes=-6,
            wloss=6,
            fed=5,
            blane=-5,
            dane=-7,
            alcohol=-3,
            drug=-5,
            psycho=-4,
            depre=-3,
        )



"""Top-level package for comorbidipy."""

__author__ = """vvcb"""
__email__ = 'vvcb.n1@gmail.com'
__version__ = '0.4.4'

from .calculator import comorbidity, hfrs, disability
from .colnames import get_colnames


#!/usr/bin/env python

"""Tests for `comorbidipy` package."""


import random
import unittest
from itertools import cycle

import pandas as pd

from comorbidipy import comorbidity, mapping

def generate_synthetic_icd10_data(
    score: str = "charlson",
    icd: str = "icd10",
    variant: str = "shmi",
    weighting: str = None,  # for compatibility
):

    mapping_key = f"{score}_{icd}_{variant}"
    patients = []

    num_of_patients = max(map(len, mapping.mapping[mapping_key].values()))
    codes = [cycle(random.sample(v, len(v))) for v in mapping.mapping[mapping_key].values()]

    for n in range(num_of_patients):

        age = random.randint(0, 100)

        for c in codes:

            patients.append({"id": n, "age": age, "code": next(c)})

    df = pd.DataFrame.from_records(patients)
    return df


class TestComorbidipy(unittest.TestCase):
    """Tests for `comorbidipy` package."""

    def setUp(self):
        """Set up test fixtures, if any."""

        self.test_args = [
            {
                "args": {
                    "icd": "icd10",
                    "score": "charlson",
                    "variant": "shmi",
                    "weighting": "shmi",
                },
                "max_ccs": 110,
            },
            {
                "args": {
                    "icd": "icd10",
                    "score": "charlson",
                    "variant": "quan",
                    "weighting": "charlson",
                },
                "max_ccs": 29,
            },
            {
                "args": {
                    "icd": "icd10",
                    "score": "charlson",
                    "variant": "quan",
                    "weighting": "quan",cd str
                },
                "max_ccs": 22,
            },
        ]
        self.number_of_runs = 20

    def tearDown(self):
        """Tear down test fixtures, if any."""
        pass


    def test_comorbidity_calculator(self):

        for x in self.test_args:

            kwargs = x["args"]
            max_ccs = x["max_ccs"]

            for i in range(self.number_of_runs):

                df = generate_synthetic_icd10_data(**kwargs)
                ccs = comorbidity(df, **kwargs)

                self.assertTrue(all(ccs["comorbidity_score"] == max_ccs))


if __name__ == "__main__":
    unittest.main()



"""Main module."""
import warnings
from functools import lru_cache
import pandas as pd
from pandas.core.common import SettingWithCopyWarning

warnings.simplefilter(action="ignore", category=SettingWithCopyWarning)

import math
from .mapping import mapping, hfrs_mapping, impairments
from .weights import weights
from .assignzero import assignzero
from .colnames import get_colnames


def _calculate_weighted_score(
    dfp: pd.DataFrame, param_score: str, assign0: bool, weighting: str
):

    # Create a copy of the supplied dataframe first
    df = dfp.copy()

    # if assign0 is True, set the less severe of the comorbidities to 0
    # this does not change the values in the actual columns but only affects the weighting
    # e.g the final dataset will still have true for both diabetes and diabetes with complications
    # but only diabetes with complications will be used for calculating the final score.
    if assign0:
        df = assignzero(df, param_score)

    # Get the weights into a pandas series
    w = weights[param_score][weighting]
    w = pd.Series(w)

    dfp["comorbidity_score"] = df.multiply(w).sum(axis=1)
    # if sum of weights is less than zero, set it zero (this only applies to UK SHMI)
    dfp["comorbidity_score"] = dfp["comorbidity_score"].where(
        dfp["comorbidity_score"] >= 0, 0
    )
    return dfp


def _age_adjust(dfp: pd.DataFrame, age: str):

    age_score = dfp[age].subtract(40).floordiv(10).apply(lambda x: min(max(0, x), 4))
    dfp["age_adj_comorbidity_score"] = dfp["comorbidity_score"] + age_score

    return dfp


def comorbidity(
    df: pd.DataFrame,
    id: str = "id",
    code: str = "code",
    age: str = "age",
    score: str = "charlson",
    icd: str = "icd10",
    variant: str = "quan",
    weighting: str = "quan",
    assign0: bool = True,
):

    # check the dataframe contains the required columns

    if id not in df.columns or code not in df.columns:
        raise KeyError(f"Missing column(s). Ensure column(s) {id}, {code} are present.")

    df = df.dropna(subset=[id, code])

    if age:
        if age not in df.columns:
            raise KeyError(f"Column age was assigned {age} but not found")
        dfid = df[[id, age]].drop_duplicates(subset=['id'])
    else:
        dfid = df[[id]].drop_duplicates()

    score_icd_variant = f"{score}_{icd}_{variant}"

    if score_icd_variant not in mapping.keys():
        raise KeyError(
            "Combination of score, icd and variant not found in mappings.\n"
            f"Allowed score_icd_variant combinations are {list(mapping)}"
        )

    reverse_mapping = {
        i: k
        for i in df[code].unique()
        for k, v in mapping[score_icd_variant].items()
        if i.startswith(tuple(v))
    }

    # Keep only codes that are in mapping
    df[code] = df[code].where(df[code].isin(reverse_mapping), other=None)

    df = df.dropna(subset=[code]).drop_duplicates(subset=[id, code])

    # Replace codes with mapping
    df[code] = df[code].replace(reverse_mapping)
    # The following is needed as multiple codes may map to same comorbidity
    # If there are duplicates, pivot will fail
    df = df.drop_duplicates(subset=[id, code])

    # Pivot
    df["tmp"] = 1
    dfp = df.pivot(index=id, columns=code, values="tmp").fillna(0)

    # If a particular comorbidity does not occur at all in the dataset,
    # create a column and assign 0
    colnames = get_colnames(score)
    for c in colnames:
        if c not in dfp.columns:
            dfp[c] = 0

    # Calculate weighted score
    dfp = _calculate_weighted_score(dfp, score_icd_variant, assign0, weighting)

    # Merge back into dfid, adjusting for age and calculating survival if needed
    if age:
        dfp = dfid.merge(dfp, on=id, how="left").fillna(0)
        dfp = _age_adjust(dfp, age)

        if score == "charlson" and weighting == "charlson":
            dfp[f"survival_10yr"] = dfp[f"age_adj_comorbidity_score"].apply(
                lambda x: 0.983 ** math.exp(0.9 * x)
            )
    else:
        dfp = dfid.merge(dfp, on=id, how="left").fillna(0)

    # Add metadata to dataframe before returning.
    # Helps when calling this function with different parameters and
    # outputs are all called 'comorbidity_score'!
    dfp.attrs = {
        "score": score,
        "icd": icd,
        "variant": variant,
        "weighting": weighting,
        "assign0": assign0,
    }

    return dfp


def hfrs(df: pd.DataFrame, id: str = "id", code: str = "code"):
    """Calculate Hospital Frailty Risk Score

    This is only applicable to patients who are 75 years or older.

    https://www.thelancet.com/journals/lancet/article/PIIS0140-6736(18)30668-8

    Args:
        df (pd.DataFrame): Dataframe with 2 columns named `id` and `code`
        id (str, optional): Name of column to use as `id`. Defaults to "id".
        code (str, optional): Name of column to use as `code`. Defaults to "code".

    Return:
        pd.DataFrame: Dataframe with `id` and `hfrs` values.
    """

    @lru_cache(maxsize=65536)
    def _mapper(x: str):
        try:
            x = x.lstrip()[0:3].upper()
            return x if x in hfrs_mapping else None
        except:
            return None

    if id not in df.columns or code not in df.columns:
        raise KeyError(f"Missing column(s). Ensure column(s) {id}, {code} are present.")

    # Keep only id, code columns and drop missing and duplicates first
    df = df[[id, code]].dropna().drop_duplicates()

    dfid = df[[id]].drop_duplicates()

    df[code] = df[code].apply(_mapper)

    # Drop missing and duplicates. This should leave only codes in hfrs_mapping
    df = df.dropna().drop_duplicates()

    df["hfrs"] = df[code].replace(hfrs_mapping)
    df = df.groupby(id)["hfrs"].sum().reset_index()

    # Merge back into original list of ids. Fill missing values with 0.
    df = dfid.merge(df, on=id, how="left").fillna(0)

    return df


def disability(df: pd.DataFrame, id: str = "id", code: str = "code") -> pd.DataFrame:
    """Identify disabilities and sensory impairments from ICD10 codes

    Args:
        df (pd.DataFrame): Pandas dataframe containing at least id and code columns
        id (str, optional): Name of column containing patient identifier. Defaults to "id".
        code (str, optional): Name of column containing ICD10 codes. Defaults to "code".

    Raises:
        KeyError: Error is raised if id or code columns are not present in dataframe.

    Returns:
        Pandas DataFrame: Pandas DataFrame with id and various disabilities/impairments columns coded as 0 or 1.
    """

    if id not in df.columns or code not in df.columns:
        raise KeyError(f"Missing column(s). Ensure column(s) {id}, {code} are present.")

    df = df.dropna(subset=[id, code])

    dfid = df[[id]].drop_duplicates()

    icd = df[code].unique()

    reverse_mapping = {
        i: k for i in icd for k, v in impairments.items() if i.startswith(tuple(v))
    }

    # Keep only codes that are in mapping
    df[code] = df[code].where(df[code].isin(reverse_mapping.keys()), other=None)

    df = df.dropna(subset=[code]).drop_duplicates(subset=[id, code])

    # Replace codes with mapping
    df[code] = df[code].replace(reverse_mapping)
    # not sure if this is needed but if there are duplicates,  pivot will fail
    df = df.drop_duplicates(subset=[id, code])
    df["tmp"] = 1

    # Pivot
    df = df.pivot(index=id, columns=code, values="tmp")

    # Merge back into original list of ids. Fill missing values with 0.
    df = dfid.merge(df, on=id, how="left").fillna(0)

    return df
